@file:Suppress("DeprecatedCallableAddReplaceWith")

package ru.dbotthepony.mc.otm.network.synchronizer

import it.unimi.dsi.fastutil.booleans.BooleanConsumer
import it.unimi.dsi.fastutil.floats.FloatConsumer
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.Reference2ObjectFunction
import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ReferenceArraySet
import mekanism.api.functions.FloatSupplier
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.core.collect.ProxiedMap
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.util.BigDecimalValueCodec
import ru.dbotthepony.mc.otm.core.util.BinaryStringCodec
import ru.dbotthepony.mc.otm.core.util.ByteValueCodec
import ru.dbotthepony.mc.otm.core.util.DecimalValueCodec
import ru.dbotthepony.mc.otm.core.util.EnumValueCodec
import ru.dbotthepony.mc.otm.core.util.IStreamCodec
import ru.dbotthepony.mc.otm.core.util.ItemStackValueCodec
import ru.dbotthepony.mc.otm.core.util.ShortValueCodec
import ru.dbotthepony.mc.otm.core.util.UUIDValueCodec
import ru.dbotthepony.mc.otm.core.util.readVarIntLE
import ru.dbotthepony.mc.otm.core.util.readVarLongLE
import ru.dbotthepony.mc.otm.core.util.writeVarIntLE
import ru.dbotthepony.mc.otm.core.util.writeVarLongLE
import ru.dbotthepony.mc.otm.secondTime
import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.InputStream
import java.lang.ref.WeakReference
import java.math.BigDecimal
import java.util.*
import java.util.function.BooleanSupplier
import java.util.function.DoubleConsumer
import java.util.function.DoubleSupplier
import java.util.function.IntConsumer
import java.util.function.IntSupplier
import java.util.function.LongConsumer
import java.util.function.LongSupplier
import java.util.function.Supplier
import kotlin.reflect.KMutableProperty0
import kotlin.reflect.KProperty
import kotlin.reflect.KProperty0

/**
 * Universal, one-to-many value synchronizer, allowing to synchronize values from server to client
 * anywhere, where input/output streams are supported
 */
@Suppress("unused")
class FieldSynchronizer(private val callback: Runnable, private val alwaysCallCallback: Boolean) {
	constructor() : this(Runnable {}, false)
	constructor(callback: Runnable) : this(callback, false)

	private var freeSlots = 0
	// почему не удалять поля напрямую?
	// чтоб не возникло проблем в состоянии гонки
	// формируем пакет -> удаляем поле по обе стороны -> клиент принимает пакет -> клиент считывает неверные данные
	// конечно, всё равно всё сломается если было удалено поле, которое находится в пакете
	// но если поля нет в пакете, то всё окей
	private val fields = ArrayList<AbstractField<*>?>(0)
	private val observers = ArrayList<AbstractField<*>>(0)

	private var nextFieldID = 0

	val hasObservers: Boolean get() = observers.isNotEmpty()
	val isEmpty: Boolean get() = fields.isEmpty()
	val isNotEmpty: Boolean get() = fields.isNotEmpty()

	var hasChanges: Boolean = false
		private set(value) {
			if (value != field) {
				field = value

				if (value && !alwaysCallCallback) {
					callback.run()
				}
			}

			if (alwaysCallCallback && value) {
				callback.run()
			}
		}

	fun markClean() {
		hasChanges = false
	}

	fun byte(getter: () -> Byte) = ComputedField(getter, ByteValueCodec)
	fun bool(getter: BooleanSupplier) = ComputedBooleanField(getter)
	fun short(getter: () -> Short) = ComputedField(getter, ShortValueCodec)
	fun long(getter: LongSupplier) = ComputedLongField(getter)
	fun fixedLong(getter: LongSupplier) = ComputedFixedLongField(getter)
	fun float(getter: FloatSupplier) = ComputedFloatField(getter)
	fun double(getter: DoubleSupplier) = ComputedDoubleField(getter)
	fun uuid(getter: () -> UUID) = ComputedField(getter, UUIDValueCodec)
	fun int(getter: IntSupplier) = ComputedIntField(getter)
	fun fixedInt(getter: IntSupplier) = ComputedFixedIntField(getter)
	fun fraction(getter: () -> Decimal) = ComputedField(getter, DecimalValueCodec)
	fun bigDecimal(getter: () -> BigDecimal) = ComputedField(getter, BigDecimalValueCodec)
	fun item(getter: () -> ItemStack) = ComputedField(getter, ItemStackValueCodec)
	fun string(getter: () -> String) = ComputedField(getter, BinaryStringCodec)

	fun byte(getter: KProperty0<Byte>) = ComputedField(getter, ByteValueCodec)
	fun bool(getter: KProperty0<Boolean>) = ComputedBooleanField(getter::get)
	fun short(getter: KProperty0<Short>) = ComputedField(getter, ShortValueCodec)
	fun long(getter: KProperty0<Long>) = ComputedLongField(getter::get)
	fun fixedLong(getter: KProperty0<Long>) = ComputedFixedLongField(getter::get)
	fun float(getter: KProperty0<Float>) = ComputedFloatField(getter::get)
	fun double(getter: KProperty0<Double>) = ComputedDoubleField(getter::get)
	fun uuid(getter: KProperty0<UUID>) = ComputedField(getter, UUIDValueCodec)
	fun int(getter: KProperty0<Int>) = ComputedIntField(getter::get)
	fun fixedInt(getter: KProperty0<Int>) = ComputedFixedIntField(getter::get)
	fun fraction(getter: KProperty0<Decimal>) = ComputedField(getter, DecimalValueCodec)
	fun bigDecimal(getter: KProperty0<BigDecimal>) = ComputedField(getter, BigDecimalValueCodec)
	fun item(getter: KProperty0<ItemStack>) = ComputedField(getter, ItemStackValueCodec)
	fun string(getter: KProperty0<String>) = ComputedField(getter, BinaryStringCodec)

	fun byte(getter: Supplier<Byte>) = ComputedField(getter::get, ByteValueCodec)
	fun bool(getter: Supplier<Boolean>) = ComputedBooleanField(getter::get)
	fun short(getter: Supplier<Short>) = ComputedField(getter::get, ShortValueCodec)
	fun long(getter: Supplier<Long>) = ComputedLongField(getter::get)
	fun fixedLong(getter: Supplier<Long>) = ComputedFixedLongField(getter::get)
	fun float(getter: Supplier<Float>) = ComputedFloatField(getter::get)
	fun double(getter: Supplier<Double>) = ComputedDoubleField(getter::get)
	fun uuid(getter: Supplier<UUID>) = ComputedField(getter::get, UUIDValueCodec)
	fun int(getter: Supplier<Int>) = ComputedIntField(getter::get)
	fun fixedInt(getter: Supplier<Int>) = ComputedFixedIntField(getter::get)
	fun fraction(getter: Supplier<Decimal>) = ComputedField(getter::get, DecimalValueCodec)
	fun bigDecimal(getter: Supplier<BigDecimal>) = ComputedField(getter::get, BigDecimalValueCodec)
	fun item(getter: Supplier<ItemStack>) = ComputedField(getter::get, ItemStackValueCodec)
	fun string(getter: Supplier<String>) = ComputedField(getter::get, BinaryStringCodec)

	fun <T : Enum<T>> enum(type: Class<T>, getter: () -> T) = ComputedField(getter, EnumValueCodec(type))
	inline fun <reified T : Enum<T>> enum(noinline getter: () -> T) = ComputedField(getter, EnumValueCodec(T::class.java))

	fun <T : Enum<T>> enum(type: Class<T>, getter: KProperty0<T>) = ComputedField(getter, EnumValueCodec(type))
	inline fun <reified T : Enum<T>> enum(getter: KProperty0<T>) = ComputedField(getter, EnumValueCodec(T::class.java))

	fun <T : Enum<T>> enum(type: Class<T>, getter: Supplier<T>) = ComputedField(getter::get, EnumValueCodec(type))
	inline fun <reified T : Enum<T>> enum(getter: Supplier<T>) = ComputedField(getter::get, EnumValueCodec(T::class.java))

	@JvmOverloads
	fun byte(
		value: Byte = 0,
		getter: FieldGetter<Byte>? = null,
		setter: FieldSetter<Byte>? = null,
	): Field<Byte> {
		return Field(value, ByteValueCodec, getter, setter)
	}

	@JvmOverloads
	fun bool(
		value: Boolean = false,
		getter: BooleanFieldGetter? = null,
		setter: BooleanFieldSetter? = null,
	): BooleanField {
		return BooleanField(value, getter, setter)
	}

	@JvmOverloads
	fun short(
		value: Short = 0,
		getter: FieldGetter<Short>? = null,
		setter: FieldSetter<Short>? = null,
	): Field<Short> {
		return Field(value, ShortValueCodec, getter, setter)
	}

	@JvmOverloads
	fun long(
		value: Long = 0L,
		getter: LongFieldGetter? = null,
		setter: LongFieldSetter? = null,
	): LongField {
		return LongField(value, getter, setter)
	}

	@JvmOverloads
	fun fixedLong(
		value: Long = 0L,
		getter: LongFieldGetter? = null,
		setter: LongFieldSetter? = null,
	): FixedLongField {
		return FixedLongField(value, getter, setter)
	}

	@JvmOverloads
	fun float(
		value: Float = 0f,
		getter: FloatFieldGetter? = null,
		setter: FloatFieldSetter? = null,
	): FloatField {
		return FloatField(value, getter, setter)
	}

	@JvmOverloads
	fun double(
		value: Double = 0.0,
		getter: DoubleFieldGetter? = null,
		setter: DoubleFieldSetter? = null,
	): DoubleField {
		return DoubleField(value, getter, setter)
	}

	@JvmOverloads
	fun uuid(
		value: UUID = UUID(0L, 0L),
		getter: FieldGetter<UUID>? = null,
		setter: FieldSetter<UUID>? = null,
	): Field<UUID> {
		return Field(value, UUIDValueCodec, getter, setter)
	}

	@JvmOverloads
	fun int(
		value: Int = 0,
		getter: IntFieldGetter? = null,
		setter: IntFieldSetter? = null,
	): IntField {
		return IntField(value, getter, setter)
	}

	@JvmOverloads
	fun string(
		value: String = "",
		getter: FieldGetter<String>? = null,
		setter: FieldSetter<String>? = null,
	): Field<String> {
		return Field(value, BinaryStringCodec, getter, setter)
	}

	@JvmOverloads
	fun fixedInt(
		value: Int = 0,
		getter: IntFieldGetter? = null,
		setter: IntFieldSetter? = null,
	): FixedIntField {
		return FixedIntField(value, getter, setter)
	}

	@JvmOverloads
	fun fraction(
		value: Decimal = Decimal.ZERO,
		getter: FieldGetter<Decimal>? = null,
		setter: FieldSetter<Decimal>? = null,
	): Field<Decimal> {
		return Field(value, DecimalValueCodec, getter, setter)
	}

	@JvmOverloads
	fun bigDecimal(
		value: BigDecimal = BigDecimal.ZERO,
		getter: FieldGetter<BigDecimal>? = null,
		setter: FieldSetter<BigDecimal>? = null,
	): Field<BigDecimal> {
		return Field(value, BigDecimalValueCodec, getter, setter)
	}

	@JvmOverloads
	fun <T : Enum<T>> enum(
		type: Class<T>,
		value: T = type.enumConstants[0],
		getter: FieldGetter<T>? = null,
		setter: FieldSetter<T>? = null,
	): Field<T> {
		return Field(value, EnumValueCodec(type), getter, setter)
	}

	@JvmOverloads
	fun <T : Enum<T>> enum(
		value: T,
		getter: FieldGetter<T>? = null,
		setter: FieldSetter<T>? = null,
	): Field<T> {
		return Field(value, EnumValueCodec(value::class.java), getter, setter)
	}

	@JvmOverloads
	fun item(
		value: ItemStack = ItemStack.EMPTY,
		getter: FieldGetter<ItemStack>? = null,
		setter: FieldSetter<ItemStack>? = null,
		observe: Boolean = true,
	): Field<ItemStack> {
		return Field(value, ItemStackValueCodec, getter, setter, isObserver = observe)
	}

	fun item(
		delegate: KMutableProperty0<ItemStack>,
	): ObservedField<ItemStack> {
		return ObservedField(delegate, ItemStackValueCodec)
	}

	private var endpointsMaxCapacity = 1
	private val endpoints = ArrayList<WeakReference<Endpoint>>(1)
	val defaultEndpoint = Endpoint()

	private var nextEndpointsCleanup = secondTime

	private fun notifyEndpoints(dirtyField: AbstractField<*>) {
		hasChanges = true

		forEachEndpoint {
			it.addDirtyField(dirtyField)
		}
	}

	private inline fun forEachEndpoint(execute: (Endpoint) -> Unit) {
		nextEndpointsCleanup = secondTime + 60

		synchronized(endpoints) {
			endpoints.forValidRefs { execute.invoke(it) }

			if (endpoints.size < endpointsMaxCapacity / 2) {
				endpoints.trimToSize()
				endpointsMaxCapacity = endpoints.size
			}
		}
	}

	inner class Endpoint {
		init {
			synchronized(endpoints) {
				endpoints.add(WeakReference(this))
				endpointsMaxCapacity = endpointsMaxCapacity.coerceAtLeast(endpoints.size)

				if (secondTime >= nextEndpointsCleanup) {
					nextEndpointsCleanup = secondTime + 60

					val iterator = endpoints.listIterator()

					for (value in iterator) {
						if (value.get() == null) {
							iterator.remove()
						}
					}

					if (endpoints.size < endpointsMaxCapacity / 2) {
						endpoints.trimToSize()
						endpointsMaxCapacity = endpoints.size
					}
				}
			}
		}

		private val dirtyFields = ReferenceArraySet<AbstractField<*>>(fields.size)

		// use LinkedList because it is ensured memory is freed on LinkedList#clear
		private val mapBacklogs = Reference2ObjectOpenHashMap<Map<*, *>, LinkedList<Pair<Any?, (DataOutputStream) -> Unit>>>()

		var unused: Boolean = false
			private set

		private var mappingVersion = -1

		fun invalidateMappings() {
			mappingVersion = -1
		}

		fun markUnused() {
			require(this === defaultEndpoint) { "This is not a default endpoint" }
			if (unused) return
			unused = true
			mapBacklogs.clear()
			dirtyFields.clear()

			val iterator = endpoints.listIterator()

			for (value in iterator) {
				if (value.get() === this) {
					iterator.remove()
				}
			}
		}

		init {
			markDirty()
		}

		fun markDirty() {
			for (field in fields) {
				field?.markDirty(this)
			}
		}

		internal fun addDirtyField(field: AbstractField<*>) {
			if (unused) {
				return
			}

			dirtyFields.add(field)
		}

		internal fun removeDirtyField(field: AbstractField<*>) {
			dirtyFields.remove(field)
		}

		internal fun <K, V> getMapBacklog(map: Map<K, V>): LinkedList<Pair<Any?, (DataOutputStream) -> Unit>> {
			if (unused) {
				return LinkedList()
			}

			return mapBacklogs.computeIfAbsent(map, Reference2ObjectFunction {
				LinkedList()
			})
		}

		fun collectNetworkPayload(): FastByteArrayOutputStream? {
			if (unused || dirtyFields.isEmpty()) {
				return null
			}

			val stream = FastByteArrayOutputStream()
			val dataStream = DataOutputStream(stream)

			for (field in dirtyFields) {
				stream.writeVarIntLE(field.id)
				field.write(dataStream, this)
			}

			dirtyFields.clear()
			stream.write(0)

			return stream
		}
	}

	private val boundEndpoints = WeakHashMap<Any, Endpoint>()

	fun computeEndpointFor(obj: Any): Endpoint {
		return boundEndpoints.computeIfAbsent(obj) { Endpoint() }
	}

	fun removeEndpointFor(obj: Any): Endpoint? {
		return boundEndpoints.remove(obj)
	}

	fun endpointFor(obj: Any): Endpoint? {
		return boundEndpoints[obj]
	}

	@Suppress("LeakingThis")
	abstract inner class AbstractField<V> : IField<V> {
		val id: Int

		init {
			if (freeSlots > 0) {
				var found = -1

				for (i in fields.indices) {
					if (fields[i] == null) {
						fields[i] = this
						found = i + 1
						freeSlots--
						break
					}
				}

				if (found == -1) {
					throw RuntimeException("freeSlots = $freeSlots but no null entries in field list!")
				} else {
					id = found
				}
			} else {
				fields.add(this)
				id = fields.size
			}
		}

		final override var isRemoved = false
			private set

		override fun remove() {
			if (isRemoved)
				return

			isRemoved = true
			freeSlots++
			fields[id - 1] = null
			observers.remove(this)

			while (fields[fields.size - 1] == null) {
				fields.removeAt(fields.size - 1)
				freeSlots--
			}

			forEachEndpoint {
				it.removeDirtyField(this)
			}
		}

		override fun markDirty(endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			endpoint.addDirtyField(this)
		}
	}

	/**
	 * Networked variable with backing field holding immutable value
	 */
	inner class Field<V>(
		private var field: V,
		private val codec: IStreamCodec<V>,
		private val getter: FieldGetter<V>? = null,
		private val setter: FieldSetter<V>? = null,
		isObserver: Boolean = false,
	) : AbstractField<V>(), IMutableField<V> {
		private var remote: V = codec.copy(field)

		init {
			if (isObserver) {
				observers.add(this)
			}
		}

		private var isDirty = false

		private val access = object : FieldAccess<V> {
			override fun read(): V {
				return field
			}

			override fun write(value: V) {
				if (!isDirty && !codec.compare(remote, value)) {
					notifyEndpoints(this@Field)
					isDirty = true
				}

				this@Field.field = value
			}
		}

		override fun observe(): Boolean {
			check(!isRemoved) { "Field was removed" }

			if (!isDirty && !codec.compare(remote, field)) {
				notifyEndpoints(this@Field)
				isDirty = true
			}

			return isDirty
		}

		override var value: V
			get() {
				val getter = this.getter

				if (getter != null) {
					return getter.invoke(access)
				}

				return this.field
			}
			set(value) {
				val setter = this.setter

				if (setter != null) {
					setter.invoke(value, access, false)
					return
				}

				if (!isDirty && !codec.compare(remote, value)) {
					notifyEndpoints(this@Field)
					isDirty = true
				}

				this.field = value
			}

		override fun markDirty() {
			check(!isRemoved) { "Field was removed" }
			notifyEndpoints(this@Field)
			isDirty = true
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			codec.write(stream, field)
			isDirty = false
			remote = codec.copy(field)
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			val value = codec.read(stream)
			val setter = this.setter

			if (setter != null) {
				setter.invoke(value, access, true)
				return
			}

			this.field = value
		}
	}

	abstract inner class PrimitiveField<V : Any>() : AbstractField<V>() {
		protected var isDirty = false

		override fun observe(): Boolean {
			check(!isRemoved) { "Field was removed" }
			return isDirty
		}

		override fun markDirty() {
			check(!isRemoved) { "Field was removed" }
			notifyEndpoints(this)
			isDirty = true
		}
	}

	/**
	 * Type specific field, storing primitive [Float] directly
	 */
	inner class FloatField(private var field: Float, private val getter: FloatFieldGetter? = null, private val setter: FloatFieldSetter? = null) : PrimitiveField<Float>(), IMutableFloatField {
		override val property = object : IMutableFloatProperty {
			override fun getValue(thisRef: Any?, property: KProperty<*>): Float {
				return this@FloatField.float
			}

			override fun setValue(thisRef: Any?, property: KProperty<*>, value: Float) {
				this@FloatField.float = value
			}
		}

		private val access = object : FloatFieldAccess {
			override fun readFloat(): Float {
				return this@FloatField.field
			}

			override fun write(value: Float) {
				if (!isDirty && value != this@FloatField.field) {
					notifyEndpoints(this@FloatField)
					isDirty = true
				}

				this@FloatField.field = value
			}
		}

		override var float: Float
			get() {
				return getter?.invoke(access) ?: this.field
			}
			set(value) {
				val setter = this.setter

				if (setter != null) {
					setter.invoke(value, access, false)
				} else if (value != this.field) {
					if (!isDirty) {
						notifyEndpoints(this)
						isDirty = true
					}

					this.field = value
				}
			}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			stream.writeFloat(this.field)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			val value = stream.readFloat()
			val setter = this.setter

			if (setter != null) {
				setter.invoke(value, access, true)
			} else {
				this.field = value
			}
		}
	}

	/**
	 * Type specific field, storing primitive [Double] directly
	 */
	inner class DoubleField(private var field: Double, private val getter: DoubleFieldGetter? = null, private val setter: DoubleFieldSetter? = null) : PrimitiveField<Double>(), IMutableDoubleField {
		override val property = object : IMutableDoubleProperty {
			override fun getValue(thisRef: Any?, property: KProperty<*>): Double {
				return this@DoubleField.double
			}

			override fun setValue(thisRef: Any?, property: KProperty<*>, value: Double) {
				this@DoubleField.double = value
			}
		}

		private val access = object : DoubleFieldAccess {
			override fun readDouble(): Double {
				return this@DoubleField.field
			}

			override fun write(value: Double) {
				if (!isDirty && value != this@DoubleField.field) {
					notifyEndpoints(this@DoubleField)
					isDirty = true
				}

				this@DoubleField.field = value
			}
		}

		override var double: Double
			get() {
				return getter?.invoke(access) ?: this.field
			}
			set(value) {
				val setter = this.setter

				if (setter != null) {
					setter.invoke(value, access, false)
				} else if (value != this.field) {
					if (!isDirty) {
						notifyEndpoints(this)
						isDirty = true
					}

					this.field = value
				}
			}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			stream.writeDouble(this.field)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			val value = stream.readDouble()
			val setter = this.setter

			if (setter != null) {
				setter.invoke(value, access, true)
			} else {
				this.field = value
			}
		}
	}

	abstract inner class AbstractIntField(protected var field: Int, private val getter: IntFieldGetter? = null, protected val setter: IntFieldSetter? = null) : PrimitiveField<Int>(), IMutableIntField {
		final override val property = object : IMutableIntProperty {
			override fun getValue(thisRef: Any?, property: KProperty<*>): Int {
				return this@AbstractIntField.int
			}

			override fun setValue(thisRef: Any?, property: KProperty<*>, value: Int) {
				this@AbstractIntField.int = value
			}
		}

		protected val access = object : IntFieldAccess {
			override fun readInt(): Int {
				return this@AbstractIntField.field
			}

			override fun write(value: Int) {
				if (!isDirty && value != this@AbstractIntField.field) {
					notifyEndpoints(this@AbstractIntField)
					isDirty = true
				}

				this@AbstractIntField.field = value
			}
		}

		final override var int: Int
			get() {
				return getter?.invoke(access) ?: this.field
			}
			set(value) {
				val setter = this.setter

				if (setter != null) {
					setter.invoke(value, access, false)
				} else if (value != this.field) {
					if (!isDirty) {
						notifyEndpoints(this)
						isDirty = true
					}

					this.field = value
				}
			}
	}

	/**
	 * Type specific field, storing primitive [Int] directly, and networking it as variable length integer
	 */
	inner class IntField(field: Int, getter: IntFieldGetter? = null, setter: IntFieldSetter? = null) : AbstractIntField(field, getter, setter) {
		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			stream.writeVarIntLE(this.field)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			val value = stream.readVarIntLE()
			val setter = this.setter

			if (setter != null) {
				setter.invoke(value, access, true)
			} else {
				this.field = value
			}
		}
	}

	/**
	 * Type specific field, storing primitive [Int] directly, and networking it as 4 octets
	 */
	inner class FixedIntField(field: Int, getter: IntFieldGetter? = null, setter: IntFieldSetter? = null) : AbstractIntField(field, getter, setter) {
		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			stream.writeInt(this.field)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			val value = stream.readInt()
			val setter = this.setter

			if (setter != null) {
				setter.invoke(value, access, true)
			} else {
				this.field = value
			}
		}
	}

	/**
	 * Type specific field, storing primitive [Long] directly
	 */
	abstract inner class AbstractLongField(protected var field: Long, private val getter: LongFieldGetter? = null, protected val setter: LongFieldSetter? = null) : PrimitiveField<Long>(), IMutableLongField {
		final override val property = object : IMutableLongProperty {
			override fun getValue(thisRef: Any?, property: KProperty<*>): Long {
				return this@AbstractLongField.long
			}

			override fun setValue(thisRef: Any?, property: KProperty<*>, value: Long) {
				this@AbstractLongField.long = value
			}
		}

		protected val access = object : LongFieldAccess {
			override fun readLong(): Long {
				return this@AbstractLongField.field
			}

			override fun write(value: Long) {
				if (!isDirty && value != this@AbstractLongField.field) {
					notifyEndpoints(this@AbstractLongField)
					isDirty = true
				}

				this@AbstractLongField.field = value
			}
		}

		final override var long: Long
			get() {
				return getter?.invoke(access) ?: this.field
			}
			set(value) {
				val setter = this.setter

				if (setter != null) {
					setter.invoke(value, access, false)
				} else if (value != this.field) {
					if (!isDirty) {
						notifyEndpoints(this)
						isDirty = true
					}

					this.field = value
				}
			}
	}

	/**
	 * Type specific field, storing primitive [Long] directly, and networking it as variable length integer
	 */
	inner class LongField(field: Long, getter: LongFieldGetter? = null, setter: LongFieldSetter? = null) : AbstractLongField(field, getter, setter) {
		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			stream.writeVarLongLE(this.field)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			val value = stream.readVarLongLE()
			val setter = this.setter

			if (setter != null) {
				setter.invoke(value, access, true)
			} else {
				this.field = value
			}
		}
	}

	/**
	 * Type specific field, storing primitive [Long] directly, and networking it as 8 octets
	 */
	inner class FixedLongField(field: Long, getter: LongFieldGetter? = null, setter: LongFieldSetter? = null) : AbstractLongField(field, getter, setter) {
		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			stream.writeLong(this.field)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			val value = stream.readLong()
			val setter = this.setter

			if (setter != null) {
				setter.invoke(value, access, true)
			} else {
				this.field = value
			}
		}
	}

	/**
	 * Type specific field, storing primitive [Boolean] directly
	 */
	inner class BooleanField(private var field: Boolean, private val getter: BooleanFieldGetter? = null, private val setter: BooleanFieldSetter? = null) : PrimitiveField<Boolean>(), IMutableBooleanField {
		override val property = object : IMutableBooleanProperty {
			override fun getValue(thisRef: Any?, property: KProperty<*>): Boolean {
				return this@BooleanField.boolean
			}

			override fun setValue(thisRef: Any?, property: KProperty<*>, value: Boolean) {
				this@BooleanField.boolean = value
			}
		}

		private val access = object : BooleanFieldAccess {
			override fun readBoolean(): Boolean {
				return this@BooleanField.field
			}

			override fun write(value: Boolean) {
				if (!isDirty && value != this@BooleanField.field) {
					notifyEndpoints(this@BooleanField)
					isDirty = true
				}

				this@BooleanField.field = value
			}
		}

		override var boolean: Boolean
			get() {
				return getter?.invoke(access) ?: this.field
			}
			set(value) {
				val setter = this.setter

				if (setter != null) {
					setter.invoke(value, access, false)
				} else if (value != this.field) {
					if (!isDirty) {
						notifyEndpoints(this)
						isDirty = true
					}

					this.field = value
				}
			}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			stream.writeBoolean(this.field)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			val value = stream.readBoolean()
			val setter = this.setter

			if (setter != null) {
				setter.invoke(value, access, true)
			} else {
				this.field = value
			}
		}
	}

	/**
	 * Networked value with backing getter which is constantly polled
	 */
	inner class ComputedField<V>(
		private val getter: () -> V,
		private val codec: IStreamCodec<V>,
		private val observer: (new: V) -> Unit = {}
	) : AbstractField<V>(), IField<V> {
		private var remote: V? = null
		private var clientValue: V? = null
		private var isDirty = false

		init {
			observers.add(this)
		}

		override fun observe(): Boolean {
			check(!isRemoved) { "Field was removed" }

			val value = value

			if (!isDirty && (remote == null || !codec.compare(remote ?: throw ConcurrentModificationException(), value))) {
				notifyEndpoints(this)
				isDirty = true
				remote = codec.copy(value)
			}

			return isDirty
		}

		override fun markDirty() {
			check(!isRemoved) { "Field was removed" }
			notifyEndpoints(this)
			isDirty = true
		}

		override val value: V
			get() = clientValue ?: getter.invoke()

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			codec.write(stream, value)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			val newValue = codec.read(stream)
			clientValue = newValue
			observer.invoke(newValue)
		}
	}

	/**
	 * Networked value with backing getter which is constantly polled
	 *
	 * This class has concrete implementation for [Float] primitive
	 */
	inner class ComputedFloatField(private val getter: FloatSupplier, private val observer: FloatConsumer = FloatConsumer {}) : AbstractField<Float>(), IFloatField {
		private var remote: Float = 0f
		private var isRemoteSet = false
		private var clientValue: Float = 0f
		private var isClientValue = false
		private var isDirty = false

		override val property = object : IFloatProperty {
			override fun getValue(thisRef: Any?, property: KProperty<*>): Float {
				return float
			}
		}

		init {
			observers.add(this)
		}

		override fun observe(): Boolean {
			check(!isRemoved) { "Field was removed" }

			val value = getter.asFloat

			if (!isDirty && (!isRemoteSet || remote != value)) {
				notifyEndpoints(this)
				isDirty = true
				isRemoteSet = true
				remote = value
			}

			return isDirty
		}

		override fun markDirty() {
			check(!isRemoved) { "Field was removed" }
			notifyEndpoints(this)
			isDirty = true
		}

		override val float: Float
			get() {
				if (isClientValue) {
					return clientValue
				} else {
					return getter.asFloat
				}
			}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			stream.writeFloat(getter.asFloat)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			val newValue = stream.readFloat()
			clientValue = newValue
			isClientValue = true
			observer.accept(newValue)
		}
	}

	/**
	 * Networked value with backing getter which is constantly polled
	 *
	 * This class has concrete implementation for [Double] primitive
	 */
	inner class ComputedDoubleField(private val getter: DoubleSupplier, private val observer: DoubleConsumer = DoubleConsumer {}) : AbstractField<Double>(), IDoubleField {
		private var remote: Double = 0.0
		private var isRemoteSet = false
		private var clientValue: Double = 0.0
		private var isClientValue = false
		private var isDirty = false

		override val property = object : IDoubleProperty {
			override fun getValue(thisRef: Any?, property: KProperty<*>): Double {
				return double
			}
		}

		init {
			observers.add(this)
		}

		override fun observe(): Boolean {
			check(!isRemoved) { "Field was removed" }

			val value = getter.asDouble

			if (!isDirty && (!isRemoteSet || remote != value)) {
				notifyEndpoints(this)
				isDirty = true
				isRemoteSet = true
				remote = value
			}

			return isDirty
		}

		override fun markDirty() {
			check(!isRemoved) { "Field was removed" }
			notifyEndpoints(this)
			isDirty = true
		}

		override val double: Double
			get() {
				if (isClientValue) {
					return clientValue
				} else {
					return getter.asDouble
				}
			}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			stream.writeDouble(getter.asDouble)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			val newValue = stream.readDouble()
			clientValue = newValue
			isClientValue = true
			observer.accept(newValue)
		}
	}

	/**
	 * Networked value with backing getter which is constantly polled
	 *
	 * This class has concrete implementation for [Int] primitive
	 */
	abstract inner class AbstractComputedIntField(protected val getter: IntSupplier, protected val observer: IntConsumer = IntConsumer {}) : AbstractField<Int>(), IIntField {
		private var remote: Int = 0
		private var isRemoteSet = false
		protected var clientValue: Int = 0
		protected var isClientValue = false
		protected var isDirty = false

		final override val property = object : IIntProperty {
			override fun getValue(thisRef: Any?, property: KProperty<*>): Int {
				return int
			}
		}

		init {
			observers.add(this)
		}

		final override fun observe(): Boolean {
			check(!isRemoved) { "Field was removed" }

			val value = getter.asInt

			if (!isDirty && (!isRemoteSet || remote != value)) {
				notifyEndpoints(this)
				isDirty = true
				isRemoteSet = true
				remote = value
			}

			return isDirty
		}

		final override fun markDirty() {
			check(!isRemoved) { "Field was removed" }
			notifyEndpoints(this)
			isDirty = true
		}

		final override val int: Int
			get() {
				if (isClientValue) {
					return clientValue
				} else {
					return getter.asInt
				}
			}
	}

	/**
	 * Networked value with backing getter which is constantly polled
	 *
	 * This class has concrete implementation for [Int] primitive, networking it as variable length integer
	 */
	inner class ComputedIntField(getter: IntSupplier, observer: IntConsumer = IntConsumer {}) : AbstractComputedIntField(getter, observer) {
		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			stream.writeVarIntLE(getter.asInt)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			val newValue = stream.readVarIntLE()
			clientValue = newValue
			isClientValue = true
			observer.accept(newValue)
		}
	}

	/**
	 * Networked value with backing getter which is constantly polled
	 *
	 * This class has concrete implementation for [Int] primitive, networking it as 4 octets
	 */
	inner class ComputedFixedIntField(getter: IntSupplier, observer: IntConsumer = IntConsumer {}) : AbstractComputedIntField(getter, observer) {
		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			stream.writeInt(getter.asInt)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			val newValue = stream.readInt()
			clientValue = newValue
			isClientValue = true
			observer.accept(newValue)
		}
	}

	/**
	 * Networked value with backing getter which is constantly polled
	 *
	 * This class has concrete implementation for [Long] primitive
	 */
	abstract inner class AbstractComputedLongField(protected val getter: LongSupplier, protected val observer: LongConsumer = LongConsumer {}) : AbstractField<Long>(), ILongField {
		private var remote: Long = 0L
		private var isRemoteSet = false
		protected var clientValue: Long = 0L
		protected var isClientValue = false
		protected var isDirty = false

		final override val property = object : ILongProperty {
			override fun getValue(thisRef: Any?, property: KProperty<*>): Long {
				return long
			}
		}

		init {
			observers.add(this)
		}

		final override fun observe(): Boolean {
			check(!isRemoved) { "Field was removed" }

			val value = getter.asLong

			if (!isDirty && (!isRemoteSet || remote != value)) {
				notifyEndpoints(this)
				isDirty = true
				isRemoteSet = true
				remote = value
			}

			return isDirty
		}

		final override fun markDirty() {
			check(!isRemoved) { "Field was removed" }
			notifyEndpoints(this)
			isDirty = true
		}

		final override val long: Long
			get() {
				if (isClientValue) {
					return clientValue
				} else {
					return getter.asLong
				}
			}
	}

	/**
	 * Networked value with backing getter which is constantly polled
	 *
	 * This class has concrete implementation for [Long] primitive, networking it as variable length integer
	 */
	inner class ComputedLongField(getter: LongSupplier, observer: LongConsumer = LongConsumer {}) : AbstractComputedLongField(getter, observer) {
		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			stream.writeVarLongLE(getter.asLong)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			val newValue = stream.readVarLongLE()
			clientValue = newValue
			isClientValue = true
			observer.accept(newValue)
		}
	}

	/**
	 * Networked value with backing getter which is constantly polled
	 *
	 * This class has concrete implementation for [Long] primitive, networking it as 8 octets
	 */
	inner class ComputedFixedLongField(getter: LongSupplier, observer: LongConsumer = LongConsumer {}) : AbstractComputedLongField(getter, observer) {
		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			stream.writeLong(getter.asLong)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			val newValue = stream.readLong()
			clientValue = newValue
			isClientValue = true
			observer.accept(newValue)
		}
	}

	/**
	 * Networked value with backing getter which is constantly polled
	 *
	 * This class has concrete implementation for [Boolean] primitive
	 */
	inner class ComputedBooleanField(private val getter: BooleanSupplier, private val observer: BooleanConsumer = BooleanConsumer {}) : AbstractField<Boolean>(), IBooleanField {
		private var remote: Boolean = false
		private var isRemoteSet = false
		private var clientValue: Boolean = false
		private var isClientValue = false
		private var isDirty = false

		override val property = object : IBooleanProperty {
			override fun getValue(thisRef: Any?, property: KProperty<*>): Boolean {
				return boolean
			}
		}

		init {
			observers.add(this)
		}

		override fun observe(): Boolean {
			check(!isRemoved) { "Field was removed" }

			val value = getter.asBoolean

			if (!isDirty && (!isRemoteSet || remote != value)) {
				notifyEndpoints(this)
				isDirty = true
				isRemoteSet = true
				remote = value
			}

			return isDirty
		}

		override fun markDirty() {
			check(!isRemoved) { "Field was removed" }
			notifyEndpoints(this)
			isDirty = true
		}

		override val boolean: Boolean
			get() {
				if (isClientValue) {
					return clientValue
				} else {
					return getter.asBoolean
				}
			}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			stream.writeBoolean(getter.asBoolean)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			val newValue = stream.readBoolean()
			clientValue = newValue
			isClientValue = true
			observer.accept(newValue)
		}
	}

	/**
	 * Networked variable with backing field holding mutable value, which is constantly observed for changes
	 */
	inner class ObservedField<V> : AbstractField<V>, IMutableField<V> {
		private val codec: IStreamCodec<V>
		private val getter: () -> V
		private val setter: (V) -> Unit
		private var remote: V

		override var value: V
			get() = getter.invoke()
			set(value) { setter.invoke(value) }

		constructor(field: KMutableProperty0<V>, codec: IStreamCodec<V>) : super() {
			this.codec = codec
			getter = field::get
			setter = field::set
			remote = codec.copy(value)
		}

		constructor(getter: () -> V, setter: (V) -> Unit, codec: IStreamCodec<V>) : super() {
			this.codec = codec
			this.getter = getter
			this.setter = setter
			remote = codec.copy(value)
		}

		private var isDirty = false

		init {
			observers.add(this)
		}

		override fun observe(): Boolean {
			check(!isRemoved) { "Field was removed" }

			val value = value

			if (!isDirty && !codec.compare(remote, value)) {
				notifyEndpoints(this)
				isDirty = true
				remote = codec.copy(value)
			}

			return isDirty
		}

		override fun markDirty() {
			check(!isRemoved) { "Field was removed" }
			notifyEndpoints(this)
			isDirty = true
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			codec.write(stream, value)
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "Field was removed" }
			this.value = codec.read(stream)
		}
	}

	inner class Map<K, V>(
		private val keyCodec: IStreamCodec<K>,
		private val valueCodec: IStreamCodec<V>,
		private val backingMap: MutableMap<K, V>,
		private val observingBackingMap: MutableMap<K, V>? = null,
		private val callback: ((changes: Collection<MapChangeset<K, V>>) -> Unit)? = null,
	) : AbstractField<MutableMap<K, V>>(), IField<MutableMap<K, V>> {
		private var isDirty = false
		private var sentAllValues = false
		private var isRemote = false

		init {
			if (observingBackingMap != null) {
				observers.add(this)
			}
		}

		private fun pushBacklog(key: Any?, value: (DataOutputStream) -> Unit) {
			forEachEndpoint {
				val list = it.getMapBacklog(this)
				val iterator = list.listIterator()

				for (pair in iterator) {
					if (pair.first == key) {
						iterator.remove()
					}
				}

				list.addLast(key to value)
			}
		}

		private fun clearBacklog() {
			forEachEndpoint {
				it.getMapBacklog(this).clear()
			}
		}

		override fun observe(): Boolean {
			check(!isRemoved) { "Field was removed" }

			if (isRemote) {
				return false
			}

			val observingBackingMap = observingBackingMap

			if (observingBackingMap != null) {
				for ((key, value) in backingMap) {
					val remoteValue = observingBackingMap[key] ?: throw ConcurrentModificationException("Backing map of $this was modified externally, or $value missed a modification")

					if (!valueCodec.compare(value, remoteValue)) {
						val valueCopy = valueCodec.copy(value)

						pushBacklog(key) {
							it.write(MapAction.ADD.ordinal + 1)
							keyCodec.write(it, key)
							valueCodec.write(it, valueCopy)
						}

						observingBackingMap[key] = valueCopy

						if (!isDirty) {
							notifyEndpoints(this)
							isDirty = true
						}
					}
				}
			}

			return isDirty
		}

		override fun markDirty() {
			check(!isRemoved) { "Field was removed" }

			if (isRemote) {
				return
			}

			if (endpoints.isEmpty()) {
				val observingBackingMap = observingBackingMap ?: return

				for ((key, value) in backingMap)
					observingBackingMap[key] = valueCodec.copy(value)

				return
			}

			isDirty = true
			val backlogs = LinkedList<LinkedList<Pair<Any?, (DataOutputStream) -> Unit>>>()

			forEachEndpoint {
				it.addDirtyField(this)
				val value = it.getMapBacklog(this)
				backlogs.add(value)
				value.clear()
				value.add(null to ClearBacklogEntry)
			}

			for ((key, value) in backingMap) {
				val valueCopy = valueCodec.copy(value)

				val action = { it: DataOutputStream ->
					it.write(MapAction.ADD.ordinal + 1)
					keyCodec.write(it, key)
					valueCodec.write(it, valueCopy)
				}

				for (backlog in backlogs) {
					backlog.add(key to action)
				}

				observingBackingMap?.put(key, valueCopy)
			}
		}

		override fun markDirty(endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }

			if (isRemote) {
				return
			}

			val backlog = endpoint.getMapBacklog(this)

			backlog.clear()
			backlog.add(null to ClearBacklogEntry)

			for ((key, value) in backingMap) {
				val valueCopy = valueCodec.copy(value)

				backlog.add(key to {
					it.write(MapAction.ADD.ordinal + 1)
					keyCodec.write(it, key)
					valueCodec.write(it, valueCopy)
				})
			}
		}

		override val value: MutableMap<K, V> = object : ProxiedMap<K, V>(backingMap) {
			override fun onClear() {
				if (isRemote) {
					return
				}

				if (endpoints.isEmpty()) {
					return
				}

				observingBackingMap?.clear()

				forEachEndpoint { endpoint ->
					endpoint.getMapBacklog(this@Map).also {
						it.clear()
						it.add(null to ClearBacklogEntry)
					}
				}

				if (!isDirty) {
					notifyEndpoints(this@Map)
					isDirty = true
				}

				hasChanges = true
			}

			override fun onValueAdded(key: K, value: V) {
				if (isRemote) {
					return
				}

				if (endpoints.isEmpty()) {
					return
				}

				val valueCopy = valueCodec.copy(value)

				pushBacklog(key) {
					@Suppress("BlockingMethodInNonBlockingContext") // false positive
					it.write(MapAction.ADD.ordinal + 1)
					keyCodec.write(it, key)
					valueCodec.write(it, valueCopy)
				}

				observingBackingMap?.put(key, valueCopy)

				if (!isDirty) {
					notifyEndpoints(this@Map)
					isDirty = true
				}

				hasChanges = true
			}

			override fun onValueRemoved(key: K, value: V) {
				if (isRemote) {
					return
				}

				if (endpoints.isEmpty()) {
					return
				}

				val keyCopy = keyCodec.copy(key)

				pushBacklog(key) {
					@Suppress("BlockingMethodInNonBlockingContext") // false positive
					it.write(MapAction.REMOVE.ordinal + 1)
					keyCodec.write(it, keyCopy)
				}

				observingBackingMap?.remove(key)

				if (!isDirty) {
					notifyEndpoints(this@Map)
					isDirty = true
				}

				hasChanges = true
			}
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			sentAllValues = false
			isDirty = false

			val iterator = endpoint.getMapBacklog(this).listIterator()

			for (entry in iterator) {
				entry.second.invoke(stream)
				iterator.remove()
			}

			stream.write(0)
		}

		override fun read(stream: DataInputStream) {
			if (!isRemote) {
				isRemote = true
				clearBacklog()
				observingBackingMap?.clear()
			}

			isDirty = false

			val changeset = LinkedList<MapChangeset<K, V>>()
			var readAction = stream.read() - 1

			while (readAction != -1) {
				if (readAction >= MapActionList.size) {
					throw IndexOutOfBoundsException("Unknown map action with ID $readAction")
				}

				when (MapActionList[readAction]) {
					MapAction.CLEAR -> {
						backingMap.clear()
						changeset.add(ClearMapChangeset)
					}

					MapAction.ADD -> {
						val key = keyCodec.read(stream)
						val value = valueCodec.read(stream)
						backingMap[key] = value
						changeset.add(MapChangeset(MapAction.ADD, key, value))
					}

					MapAction.REMOVE -> {
						val key = keyCodec.read(stream)
						backingMap.remove(key)
						changeset.add(MapChangeset(MapAction.REMOVE, key, null))
					}
				}

				readAction = stream.read() - 1
			}

			if (changeset.size != 0) {
				callback?.invoke(changeset)
			}
		}
	}

	/**
	 * marks all fields dirty, invalidates mappings for each endpoint
	 */
	fun invalidate() {
		for (field in fields) {
			field?.markDirty()
		}

		forEachEndpoint {
			it.invalidateMappings()
		}
	}

	/**
	 * Observe changes of all fields with backing computation lambda
	 */
	fun observe(): Boolean {
		var changes = false

		if (observers.isNotEmpty()) {
			for (field in observers) {
				changes = field.observe() || changes
			}
		}

		return changes
	}

	/**
	 * [defaultEndpoint]#collectNetworkPayload
	 */
	fun collectNetworkPayload(): FastByteArrayOutputStream? {
		check(!defaultEndpoint.unused) { "Default endpoint is not used" }
		observe()
		val values = defaultEndpoint.collectNetworkPayload()
		markClean()
		return values
	}

	fun read(stream: InputStream): Int {
		var fieldId = stream.readVarIntLE()
		var i = 0
		val dataStream = DataInputStream(stream)

		while (fieldId != 0) {
			val field = fields.getOrNull(fieldId - 1) ?: throw IllegalArgumentException("Unknown field ID ${fieldId - 1}")
			field.read(dataStream)
			fieldId = stream.readVarIntLE()
			i++
		}

		return i
	}

	companion object {
		private val ClearBacklogEntry = { stream: DataOutputStream -> stream.write(MapAction.CLEAR.ordinal + 1) }
		private val MapActionList = MapAction.values()
		private val ClearMapChangeset = MapChangeset(MapAction.CLEAR, null, null)
	}
}
